/*
 	maxriss: find cracks (black lines) in pictures and measure their length.
 	Dieses Programm ließt R-Matrix-Dumps ein, erkennt dann Risse anhand eines Schwellwertes.
 	Da ein Riss eine Linie und keine Fläche (was man durch den Schwellwert bekommt) ist, 
 	wird die längste kürzeste Verbindungsstrecke zwischen zwei Punkten in der Fläche als Riss bestimmt.
 	Die berechneten Daten werden wieder in eine für R lesbare Datei geschrieben.
 	Es wird geschrieben: 
 		Die Länge des Risses
 		Die Anzahl der Pixel der Fläche
 		Startpunkt des Risses
 		Endpunkt des Risses
 		
 	Compiler-Optionen:
 	GRAY: Definiere, um Grau-Bilder zu bearbeiten. Default: Schwarz-Weiß
 	STANDALONE: Definiere, damit dieses Programm als eigenständige Applikation ausgeführt werden kann. Default: R-Extension
 		
 	Copyright (C) 23.3.2008  Arne Müller
 	
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */


#include <vector>
#include <algorithm>
#include <stack>
#include <queue>

#include <R.h>
#include <Rdefines.h>

using namespace std; 

#define RISS_DEFAULT_SCHWELLE 150
// Kosten bei der kürzeste-Riss-Suche
#define COST_VERT 1
#define COST_HORI 1
#define COST_DIAG 1.414213562
#define EPSILON 0.0001

#define PUSH(predecessor, next,d,ol) if(dists[next.x][next.y]-EPSILON > d) { \
	dists[next.x][next.y] = d; \
	optLength[next.x][next.y] = ol; \
	pd.dist = d; \
	pd.optlength = ol; \
	pd.p = next; \
	q.push(pd); \
	predecessor_x[next.x][next.y] = predecessor.x; \
	predecessor_y[next.x][next.y] = predecessor.y; \
}

	struct s_crack {
		float max_length;
		int num_punkte;
		int startx;
		int starty;
		int endex;
		int endey;
	};
	struct s_crackline {
		SEXP line; // Ist nur ein Pointer, es wird also nix boeses hin-und hergeschoben
	};
	struct s_crackcluster {
		SEXP points;  // Ist nur ein Pointer, es wird also nix boeses hin-und hergeschoben
	};
	vector<s_crack> cracks;
	vector<s_crackline> cracklines;
	vector<s_crackcluster> crackcluster;


bool** visited;
float** dists;
float** optLength;
int** predecessor_x;
int** predecessor_y;
float MAX = pow(2,31);

struct punkt{
	int x;
	int y;
	bool operator<(punkt p) const {
		return(dists[x][y] > dists[p.x][p.y]);
	}
};

struct punkt_dist {
	punkt p;
	float dist;
	float optlength;
	bool operator<(punkt_dist pd) const {
		return(dist > pd.dist);
	}
};

float** bild;
float riss_schwelle = RISS_DEFAULT_SCHWELLE;
int W;
int H;
float alpha;
float beta;
int minSize;
float max_Riss_laenge = 0;
int anz_Risse = 0;

	int num_protected = 0;

	void printCrack(float max_length, int num_punkte, punkt &start, punkt &ende) {
		s_crack crack;
		crack.max_length = max_length;
		crack.num_punkte = num_punkte;
		crack.startx = start.x+1;
		crack.starty = start.y+1;
		crack.endex = ende.x+1;
		crack.endey = ende.y+1;
		cracks.push_back(crack);
	}
	
	void printCrackLine(stack<punkt> &path) {
		s_crackline crack;
		int length = path.size();
		PROTECT(crack.line = allocVector(INTSXP,length*2));
		num_protected++;
		int* line = INTEGER_POINTER(crack.line);
		while(!path.empty()) {
			punkt p = path.top();
			path.pop();
			line[0] = p.x+1;
			line[1] = p.y+1;
			line+=2;
		}
		
		cracklines.push_back(crack);
	}
	
	void printPoints(vector<punkt> &punkte) {
		s_crackcluster crack;
		int length = punkte.size();
		PROTECT(crack.points = allocVector(INTSXP,length*2));
		num_protected++;
		int *points = INTEGER_POINTER(crack.points);
		for(int i = 0; i < length; i++) {
			points[2*i] = punkte[i].x+1;
			points[2*i+1] = punkte[i].y+1;
		}
				
		crackcluster.push_back(crack);
	}


	bool istRiss(punkt p) {
		return(bild[p.x][p.y] < riss_schwelle);
	}

	bool istRiss(int x, int y) {
		return(bild[x][y] < riss_schwelle);
	}

// Gewichtsfunktion
float gewicht(float a, float b, float abstand)   {
	return abstand*(1 + alpha*(a+b) + beta*sqrt((a-b)*(a-b)));
}


/**
Berechnet Riss, der von start_x, start_y startet.
*/
void riss(int start_x, int start_y) {
	punkt start;
	start.x = start_x;
	start.y = start_y;
	if(!istRiss(start)) {
		visited[start_x][start_y] = true;
		return;
	}
	// neuer Riss gefunden
	anz_Risse++;
	// Fasst man das Bild als Graphen auf bei dem die Pixel Knoten sind und Schwarze-Riss-Pixel verbunden sind,
	// so muss um die Länge der Risse zu bestimmen die Zusammenhangskomponenten des Graphen gefunden werden
	// Dies ist über eine Tiefensuche möglich
	// Die Punkte (Pixel) des Risses werden in punkte gespeichert
	vector<punkt> punkte;
	stack<punkt> to_test;
	to_test.push(start);
	
	while(!to_test.empty()) {
		punkt p = to_test.top();
		to_test.pop();
		if((!visited[p.x][p.y]) && istRiss(p)) {
			// Punkt ist teil des Risses
			punkte.push_back(p);
			punkt next;
			next.x = p.x+1;
			next.y = p.y;
			if(next.x < W)                 to_test.push(next);
			next.y = p.y+1;
			if(next.x < W &&  next.y < H)  to_test.push(next);
			next.x = p.x;
			if(               next.y < H)  to_test.push(next);
			next.x = p.x-1;
			if(next.x >= 0 && next.y < H)  to_test.push(next);
			next.y = p.y;
			if(next.x >= 0)                to_test.push(next);
			next.y = p.y-1;
			if(next.x >= 0 && next.y >= 0) to_test.push(next);
			next.x = p.x;
			if(               next.y >= 0) to_test.push(next);
			next.x = p.x+1;
			if(next.x < W &&  next.y >= 0) to_test.push(next);
		}
		visited[p.x][p.y] = true;
	}
	
	if(punkte.size() > minSize)   {
		printPoints(punkte);
		
		// Rprintf("found crack with size: %i\n", punkte.size());
		
		// alle Punkte des Risses liegen nun in punkte
		
		// Führt man an jedem Punkt eine Kürzeste-Wege-Suche durch, erhält man das gleiche Resultat in O(n^2 log n)!
		// Breitensuche geht nicht, da diagonalen mit WURZEL(2) gewichtet werden sollen und somit alle Kanten nicht das gleiche Gewicht haben.
		
		// Idee: 
		// Da eigentlich nur lokale Maxima (von einer Breitensuche von einem beliebigen Punkt) mögliche Start bzw. Endpunkte sind,
		// lässt sich damit die Laufzeit noch weiter reduzieren.
		// Beweis?
		// -> noch nicht implementiert
		
		float max_length = -1;
		float max_dist = -1;
		punkt lstart;
		punkt lende;
		stack<punkt> path;
		
		for(unsigned int i = 0; i < punkte.size(); i++) {
			// dist-Tabelle für Wichtige bereiche initialisieren
			// alte Reste müssen nicht gelöscht werden, da diese ehh disjunkt sind
			
			for(unsigned int j = 0; j < punkte.size(); j++) {
				dists[punkte[j].x][punkte[j].y] = MAX;
				optLength[punkte[j].x][punkte[j].y] = MAX;
				predecessor_x[punkte[j].x][punkte[j].y] = -1;
				predecessor_y[punkte[j].x][punkte[j].y] = -1;
			}
			dists[punkte[i].x][punkte[i].y] = 0;
			priority_queue<punkt_dist> q;
			punkt_dist pd;
			pd.p = punkte[i];
			pd.dist = 0;
			pd.optlength = 0;
			q.push(pd);
			while(!q.empty()) {
				punkt_dist pd = q.top();
				punkt p = pd.p;
				q.pop();
				
				if(pd.dist < dists[p.x][p.y]+EPSILON) {
					float dist = pd.dist;
					float opt_length = pd.optlength;
					if(max_length < opt_length || (max_length == opt_length && dist < max_dist)) {
						lstart = punkte[i];
						lende = p;
						max_length = opt_length;
						max_dist = dist;
					}
					punkt next;
					next.x = p.x+1;
					next.y = p.y;
					if(next.x < W && istRiss(next))                 PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_HORI), opt_length+COST_HORI);
					next.y = p.y+1;
					if(next.x < W &&  next.y < H && istRiss(next))  PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_DIAG), opt_length+COST_DIAG);
					next.x = p.x;
					if(               next.y < H && istRiss(next))  PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_VERT), opt_length+COST_VERT);
					next.x = p.x-1;
					if(next.x >= 0 && next.y < H && istRiss(next))  PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_DIAG), opt_length+COST_DIAG);
					next.y = p.y;
					if(next.x >= 0 && istRiss(next))                PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_HORI), opt_length+COST_HORI);
					next.y = p.y-1;
					if(next.x >= 0 && next.y >= 0 && istRiss(next)) PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_DIAG), opt_length+COST_DIAG);
					next.x = p.x;
					if(               next.y >= 0 && istRiss(next)) PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_VERT), opt_length+COST_VERT);
					next.x = p.x+1;
					if(next.x < W &&  next.y >= 0 && istRiss(next)) PUSH(p, next, dist+gewicht(bild[p.x][p.y], bild[next.x][next.y],COST_DIAG), opt_length+COST_DIAG);
				}
			}
			if((lstart.x == punkte[i].x) && (lstart.y == punkte[i].y)) {
				// Von diesem Startpunkt wurde ein längerer Pfad gefunden
				while(!path.empty()) path.pop(); // path clearen
				punkt p = lende;
				while((p.x != lstart.x) || (p.y != lstart.y)) {
					path.push(p);
					
					punkt aktuellerPunkt = p;
					p.x = predecessor_x[aktuellerPunkt.x][aktuellerPunkt.y];
					p.y = predecessor_y[aktuellerPunkt.x][aktuellerPunkt.y];
				}
				path.push(lstart);
			}
		}
		
		printCrack(max_length, punkte.size(), lstart, lende);
		printCrackLine(path);
		
		max_Riss_laenge = max(max_length, max_Riss_laenge);
	}
}

int findRisse() {
	max_Riss_laenge = 0;
	anz_Risse = 0;
	//visited initialisieren
	visited = new bool*[W];
	dists = new float*[W];
	optLength = new float*[W];
	predecessor_x = new int*[W];
	predecessor_y = new int*[W];
	for(int x = 0; x < W; x++) {
		visited[x] = new bool[H];
		dists[x] = new float[H];
		optLength[x]= new float[H];
		predecessor_x[x] = new int[H];
		predecessor_y[x] = new int[H];
		for(int y = 0; y < H; y++) {
			visited[x][y] = false;
			dists[x][y] = 0;
			optLength[x][y] = 0;
			predecessor_x[x][y] = -1;
			predecessor_y[x][y] = -1;
		}
	}
	
	for(int x = 0; x < W; x++) {
		for(int y = 0; y < H; y++) {
			if(!visited[x][y]) {
				riss(x, y);
			}
		}
	}
	for(int x = 0; x < W; x++) {
		delete [] visited[x];
		delete [] dists[x];
		delete [] optLength[x];
		delete [] predecessor_x[x];
		delete [] predecessor_y[x];
	}
	delete[] visited;
	delete[] dists;
	delete [] optLength;
	delete [] predecessor_x;
	delete [] predecessor_y;
	
	Rprintf("max. crack length: %f\n", max_Riss_laenge);
	return(max_Riss_laenge);
}



	SEXP maxriss2(SEXP r_bild, SEXP r_breite, SEXP r_schwellenwert, SEXP r_alpha, SEXP r_beta, SEXP r_minsize) {
		riss_schwelle = NUMERIC_VALUE(r_schwellenwert);
		alpha = NUMERIC_VALUE(r_alpha);
		beta = NUMERIC_VALUE(r_beta);
		minSize = INTEGER_VALUE(r_minsize);
			double *pixel;
		
			W=INTEGER_VALUE(r_breite);
			pixel = NUMERIC_POINTER(r_bild);
			H = length(r_bild)/W;

			// Bilddaten in Matrix umformen
			bild = new float*[W];
			for(int x = 0; x < W; x++) {
				bild[x] = new float[H];
				for(int y = 0; y < H; y++) {
					bild[x][y] = pixel[y*W+x];
				}
			}

			//von alten Ergebnissen säubern
			cracks.clear();
			cracklines.clear();
			crackcluster.clear();
			num_protected = 0;
		
			// Risse finden
			//Rprintf("berechne Risse von Bild mit Dimension %i x %i\n", W, H);

			findRisse();

			// Erzeuge nun die R-Strukturen für die Ausgabe
			// Cracklängen Liste
			Rprintf("%i cracks found\n", cracks.size());
			SEXP crack;
			PROTECT(crack = allocVector(REALSXP,cracks.size()*6));
			num_protected++;
			double *cc = NUMERIC_POINTER(crack);
			for(int i = 0; i < cracks.size(); i++) {
				// hinzufügen
				s_crack c = cracks[i];
				cc[0] = c.max_length;
				cc[1] = c.num_punkte;
				cc[2] = c.startx;
				cc[3] = c.starty;
				cc[4] = c.endex;
				cc[5] = c.endey;
				cc+=6;
			}			
			
			// Cracklinien
			SEXP line;
			PROTECT(line = allocVector(VECSXP,cracklines.size()));
			num_protected++;
			for(int i = 0; i < cracklines.size(); i++) {
				SET_VECTOR_ELT(line, i, cracklines[i].line);
			}

			// Punkte der Cracks
			SEXP cluster;
			PROTECT(cluster = allocVector(VECSXP,crackcluster.size()));
			num_protected++;
			for(int i = 0; i < crackcluster.size(); i++) {
				SET_VECTOR_ELT(cluster, i, crackcluster[i].points);
			}
			
			// Liste der Listen TODO ist das so wirklich richtig?
			SEXP list_names;
			PROTECT(list_names = allocVector(STRSXP, 3));
			num_protected++;
			SET_STRING_ELT(list_names, 0,  mkChar("crackpaths"));
			SET_STRING_ELT(list_names, 1,  mkChar("cracks"));
			SET_STRING_ELT(list_names, 2,  mkChar("crackclusters"));
			
			SEXP list;
			PROTECT(list = allocVector(VECSXP,3));
			num_protected++;
			SET_VECTOR_ELT(list, 0, line);
			SET_VECTOR_ELT(list, 1, crack);
			SET_VECTOR_ELT(list, 2, cluster);
			setAttrib(list, R_NamesSymbol, list_names);
		
			for(int x = 0; x < W; x++) {
				delete [] bild[x];
			}
			delete [] bild;
			UNPROTECT(num_protected);
			return(list);
	}

	extern "C" {
		SEXP crackrec(SEXP bild, SEXP breite, SEXP schwelle, SEXP alpha, SEXP beta, SEXP minsize) {
			return maxriss2(bild,breite,schwelle, alpha, beta, minsize);
		}
	}

